/*
 *     SocialLedge.com - Copyright (C) 2013
 *
 *     This file is part of free software framework for embedded processors.
 *     You can use it and/or distribute it as long as this copyright header
 *     remains unmodified.  The code is free for personal use and requires
 *     permission to use in a commercial product.
 *
 *      THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
 *      OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 *      MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
 *      I SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
 *      CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 *     You can reach the author of this software at :
 *          p r e e t . w i k i @ g m a i l . c o m
 */

/**
 * @file
 * @brief This is the application entry point.
 * 			FreeRTOS and stdio printf is pre-configured to use uart0_min.h before main() enters.
 * 			@see L0_LowLevel/lpc_sys.h if you wish to override printf/scanf functions.
 *
 */

#include "tasks.hpp"
#include "LPC17xx.h"
#include <stdio.h>
#include <integer.h>
#include <cstring>
#include "uart0_min.h"
#include "storage.hpp"
#include "NRFWireless.h"
#include "sta013.h"
#include "ssp1.h"
#include "spi_sem.h"

#include "WirelessConfig.h"

class ReaderTask;
class WriterTask;
class Mp3Task;

#define NRF_PAYLOAD 32

#define PACKET_HEADER_SIZE 1

#define FILE_HEADER_SIZE 2

const unsigned int FILE_CHUNK_SIZE = 4096;

const unsigned int MAX_DATA_PACKET = (NRF_PAYLOAD - PACKET_HEADER_SIZE);

const unsigned int QUEUE_ITENS = 2;
const unsigned int QUEUE_ITEM_SIZE = FILE_CHUNK_SIZE + FILE_HEADER_SIZE;

const char* FILENAME = "1:artic.mp3";

typedef enum
{
	DATA = 0x00,
	PLAY,
	PAUSE,
	STOP,
	MUTE,
	END,
	MAX_CMD = 0x08 // no more than 3 bits
} header_cmd_t;

QueueHandle_t g_sender_queue;
QueueHandle_t g_receiver_queue;

Mp3Task * mp3Task = NULL;

void buildPacket(char* packet, header_cmd_t cmd, char* buffer, unsigned int dataSize)
{
	if(dataSize > MAX_DATA_PACKET)
	{
		uart0_puts("Wrong data size");
		return;
	}

	//first 3 bits are commands, next 5 bits are the actual packet size.
	packet[0] = (cmd << 5);
	packet[0] |= dataSize;

	if(buffer)
	{
		memcpy(packet + PACKET_HEADER_SIZE, buffer, dataSize);
	}
}

class ReaderTask : public scheduler_task
{
private:
	char data[QUEUE_ITEM_SIZE] = {0};
	unsigned int offset = 0;
	FRESULT status;

public:
	ReaderTask(uint8_t priority) :
		scheduler_task("Reader", QUEUE_ITEM_SIZE + 1024, priority)
	{
	}

	bool run(void *p)
	{
		unsigned int bytesRead = 0;

		char option = 0;

	    printf("\n1 - Send file");
	    printf("\n2 - Play");
	    printf("\n3 - Pause");
	    printf("\n4 - Stop");
	    printf("\n5 - Mute 1");
	    printf("\n6 - Mute 2\n");
	    option = uart0_getchar(option);

	    if(option == '1')
	    {
	    	uart0_puts("Sending song...");
	    	offset = 0;
	    	for (;;)
			{
	    		status = Storage::myRead(FILENAME, &data[FILE_HEADER_SIZE], FILE_CHUNK_SIZE, offset, &bytesRead);
				if(FR_OK != status || 0 == bytesRead) {
					break;
				}

				// bytes in chunk
				data[0] = (bytesRead >> 8) & 0xFF;
				data[1] = (bytesRead & 0xFF);

				xQueueSend(g_sender_queue, data, portMAX_DELAY );

				offset += bytesRead;
			}

			data[0] = data[1] = 0;
			xQueueSend(g_sender_queue, data, portMAX_DELAY );

			uart0_puts("End of file");
	    }
	    else if(option == '2')
	    {
	    	uart0_puts("Sending play");
	    	char packet[NRF_PAYLOAD];
	    	buildPacket(packet, PLAY, NULL, 0);
	    	SenderWireless_service(packet, NRF_PAYLOAD);
	    }
	    else if(option == '3')
		{
	    	uart0_puts("Sending pause.");
			char packet[NRF_PAYLOAD];
			buildPacket(packet, PAUSE, NULL, 0);
			SenderWireless_service(packet, NRF_PAYLOAD);
		}
	    else if(option == '4')
		{
			uart0_puts("Sending stop.");
			char packet[NRF_PAYLOAD];
			buildPacket(packet, STOP, NULL, 0);
			SenderWireless_service(packet, NRF_PAYLOAD);
		}
	    else if(option == '5')
		{
			uart0_puts("Sending Mute 1.");
			char speaker = 1;
			char packet[NRF_PAYLOAD];
			buildPacket(packet, MUTE, &speaker, 1);
			SenderWireless_service(packet, NRF_PAYLOAD);
		}
	    else if(option == '6')
		{
			uart0_puts("Sending Mute 2.");
			char speaker = 2;
			char packet[NRF_PAYLOAD];
			buildPacket(packet, MUTE, &speaker, 1);
			SenderWireless_service(packet, NRF_PAYLOAD);
		}

		return true;
	}

};

/**
 * Nordic wireless task for use to send MP3 in SoundBuddy
 */
class SenderWirelessTask : public scheduler_task
{
	private:
		char data[QUEUE_ITEM_SIZE] = {0};
		unsigned int totalBytesTransferred = 0;

    public:
        SenderWirelessTask(uint8_t priority) :
            scheduler_task("SenderWireless", QUEUE_ITEM_SIZE + 1024, priority)
        {
            nrf_driver_init();
        }

        bool run(void *p)
        {
        	unsigned int before = 0;
        	unsigned int timeSpent = 0;
        	char packet[NRF_PAYLOAD];

        	for(;;)
        	{
        		xQueueReceive(g_sender_queue, data, portMAX_DELAY);

        		if(before == 0)
        			before = sys_get_uptime_ms();

        		unsigned int bytesToTransfer = (data[0] << 8) | data[1];
        		char * fileDataPtr = &data[FILE_HEADER_SIZE];

        		if( 0 == bytesToTransfer ) //no bytes to transfer, end of file
        		{
        			buildPacket(packet, END, NULL, 0);
					SenderWireless_service(packet, NRF_PAYLOAD);
					break;
        		} else { // bytes to transfer

        			unsigned int packetsToTransfer = bytesToTransfer / MAX_DATA_PACKET;

					for(unsigned int i=0; i < packetsToTransfer; i++)
					{
						for(volatile int delay=0; delay < 50; delay++)
						{
							taskYIELD();
						}

						buildPacket(packet, DATA, &fileDataPtr[i*MAX_DATA_PACKET], MAX_DATA_PACKET);

						SenderWireless_service(packet, NRF_PAYLOAD);
						totalBytesTransferred += MAX_DATA_PACKET;
					}

					unsigned int unalignedBytes = bytesToTransfer % MAX_DATA_PACKET;
					if(unalignedBytes > 0) // some bytes were left off
					{
						buildPacket(packet, DATA, &fileDataPtr[bytesToTransfer - unalignedBytes], unalignedBytes);

						SenderWireless_service(packet, NRF_PAYLOAD);
						totalBytesTransferred += unalignedBytes;
					}
        		}
        	}

//        	timeSpent = sys_get_uptime_ms() - before;
//        	unsigned int kbps = totalBytesTransferred * 8 / timeSpent;
//
//			char duration[50] = {0};
//			sprintf(duration, "%u\t %u\t %u", timeSpent, totalBytesTransferred, kbps);
//
//			uart0_puts("\nTime\t bytes\t kbps:");
//			uart0_puts(duration);

			return true;
        }
};

class Mp3Task : public scheduler_task
{
	char data[FILE_CHUNK_SIZE];
	unsigned int offset = 0;
	FRESULT status;
public:
	Mp3Task(uint8_t priority) :
		scheduler_task("Mp3", QUEUE_ITEM_SIZE + 1024, priority)
	{
		initialize_sta013();
	}

	void stop()
	{
		offset = 0;
	}

	bool run(void * p)
	{
		unsigned int bytesRead = 0;
		status = Storage::myRead(FILENAME, data, sizeof(data), offset, &bytesRead);
		if(FR_OK != status || bytesRead == 0) {
			uart0_puts("No buffer");
			offset = 0;
			return true;
		}

		offset += bytesRead;

		unsigned int streamedBytes = 0;
		while(streamedBytes < bytesRead) {
			if(STA013_NEEDS_DATA())
			{
				spi1_lock();
				SELECT_MP3_CS();
				ssp1_exchange_byte(data[streamedBytes++]);
				DESELECT_MP3_CS();
				spi1_unlock();
			}
		}
	}
};

/**
 * Nordic wireless task for use to receive MP3 in SoundBuddy
 */
class ReceiverWirelessTask : public scheduler_task
{
	private:
		char dataBuffer[QUEUE_ITEM_SIZE];
		FRESULT status = FR_INT_ERR;
		FIL file;
		unsigned int offset = 0;

    public:
        ReceiverWirelessTask(uint8_t priority) :
            scheduler_task("ReceiverWireless", QUEUE_ITEM_SIZE + 1024, priority)
        {
            nrf_driver_init();
        }

        void decodePacket(char* packet, header_cmd_t* cmd, char** dataPtr, unsigned int* dataSize)
        {
        	*cmd = (header_cmd_t) ((packet[0] >> 5) & 0x07);
        	*dataSize = (packet[0] & 0x1F);
        	*dataPtr = packet + PACKET_HEADER_SIZE;
        }

        void queueBuffer(char* buffer, unsigned int bytesInBuffer)
        {
        	// bytes in chunk
			dataBuffer[0] = (bytesInBuffer >> 8) & 0xFF;
			dataBuffer[1] = (bytesInBuffer & 0xFF);

			xQueueSend(g_receiver_queue, dataBuffer, portMAX_DELAY );
			return;
        }

        bool run(void *p)
        {

        	vTaskSuspend(mp3Task->getTaskHandle());

        	while(1) {
				unsigned int before = 0;
				unsigned int timeSpent = 0;
				unsigned int totalBytesReceived = 0;
				unsigned int bytesInBuffer = 0;

				char packet[NRF_PAYLOAD] = {0};

				uart0_puts("\nReceiving...\n");
				for(;;)
				{
					ReceiverWireless_service(packet, NRF_PAYLOAD);

					if(before == 0) {
						before = sys_get_uptime_ms();
					}

					header_cmd_t cmd;
					unsigned int dataSize = 0;
					char* dataPtr;
					decodePacket(packet, &cmd, &dataPtr, &dataSize);

					totalBytesReceived += dataSize;

					if(cmd == DATA)
					{
						//Remember the offset for the Header
						memcpy(&dataBuffer[bytesInBuffer + FILE_HEADER_SIZE], dataPtr, dataSize);
						bytesInBuffer += dataSize;
					}
					else if(cmd == END)
					{
						if(0 < bytesInBuffer) //we have data in the buffer
						{
							//send the last chunk
							queueBuffer(dataBuffer, bytesInBuffer);
						}

						break;
					}
					else if(cmd == PLAY)
					{
						uart0_puts("Playing now!");
						sta013ResumeDecoder();
						vTaskResume(mp3Task->getTaskHandle());
					}
					else if(cmd == PAUSE)
					{
						uart0_puts("Pausing now!");
						vTaskSuspend(mp3Task->getTaskHandle());
					}
					else if(cmd == STOP)
					{
						uart0_puts("Stopping now!");
						mp3Task->stop();
						vTaskSuspend(mp3Task->getTaskHandle());
					}
					else if(cmd == MUTE)
					{
						if(packet[1] == SPEAKER) //Its me!
						{
							uart0_puts("Muting now!");
							sta013MuteDecoder();
						}
					}

					if((bytesInBuffer + MAX_DATA_PACKET) >= QUEUE_ITEM_SIZE) //we can't hold another packet
					{
						queueBuffer(dataBuffer, bytesInBuffer);
						//flush dataBuffer
						bytesInBuffer = 0;
					}
				}

				timeSpent = sys_get_uptime_ms() - before;
				unsigned int kbps = totalBytesReceived * 8 / timeSpent;

				char strInfo[50] = {0};
				sprintf(strInfo, "%u\t %u\t %u", timeSpent, totalBytesReceived, kbps);

				uart0_puts("\nTime\t bytes\t kbps:");
				uart0_puts(strInfo);

				//send end of file
				queueBuffer(dataBuffer, 0);
			}

            return true;
        }
};

class WriterTask : public scheduler_task
{
private:
	char data[QUEUE_ITEM_SIZE];
	FRESULT status;
	unsigned int offset = 0;

public:
	WriterTask(uint8_t priority) :
			scheduler_task("Writer", QUEUE_ITEM_SIZE + 1024, priority)
		{

		}

		bool run(void *p)
		{
			unsigned int bytesToWrite = 0;
			unsigned int bytesWritten = 0;

			xQueueReceive(g_receiver_queue, data, portMAX_DELAY);

			bytesToWrite = (data[0] << 8) | data[1];
			if(bytesToWrite == 0)
			{
				char strInfo[50] = {0};
				sprintf(strInfo, "%u", offset);

				uart0_puts("bytes - written");
				uart0_puts(strInfo);

				offset = 0;

				return true;
			}

			status = Storage::myWrite(FILENAME, data + FILE_HEADER_SIZE, bytesToWrite, offset, &bytesWritten);
			if(status != FR_OK || bytesWritten != bytesToWrite)
			{
				uart0_puts("Error writing file");
			}

			offset += bytesWritten;

			return true;
		}
};



/**
 * The main() creates tasks or "threads".  See the documentation of scheduler_task class at scheduler_task.hpp
 * for details.  There is a very simple example towards the beginning of this class's declaration.
 *
 * @warning SPI #1 bus usage notes (interfaced to SD & Flash):
 *      - You can read/write files from multiple tasks because it automatically goes through SPI semaphore.
 *      - If you are going to use the SPI Bus in a FreeRTOS task, you need to use the API at L4_IO/fat/spi_sem.h
 *
 * @warning SPI #0 usage notes (Nordic wireless)
 *      - This bus is more tricky to use because if FreeRTOS is not running, the RIT interrupt may use the bus.
 *      - If FreeRTOS is running, then wireless task may use it.
 *        In either case, you should avoid using this bus or interfacing to external components because
 *        there is no semaphore configured for this bus and it should be used exclusively by nordic wireless.
 */
int main(void)
{
    /**
     * A few basic tasks for this bare-bone system :
     *      1.  Terminal task provides gateway to interact with the board through UART terminal.
     *      2.  Remote task allows you to use remote control to interact with the board.
     *      3.  Wireless task responsible to receive, retry, and handle mesh network.
     *
     * Disable remote task if you are not using it.  Also, it needs SYS_CFG_ENABLE_TLM
     * such that it can save remote control codes to non-volatile memory.  IR remote
     * control codes can be learned by typing the "learn" terminal command.
     */
//    scheduler_add_task(new terminalTask(PRIORITY_HIGH));

#if NRF_SENDER
    g_sender_queue = xQueueCreate( QUEUE_ITENS, QUEUE_ITEM_SIZE);
    scheduler_add_task(new ReaderTask(PRIORITY_MEDIUM));
    scheduler_add_task(new SenderWirelessTask(PRIORITY_HIGH));
#else
    g_receiver_queue = xQueueCreate(QUEUE_ITENS, QUEUE_ITEM_SIZE);
    scheduler_add_task(new WriterTask(PRIORITY_LOW));
    mp3Task = new Mp3Task(PRIORITY_MEDIUM);
    scheduler_add_task(mp3Task);
    scheduler_add_task(new ReceiverWirelessTask(PRIORITY_HIGH));
#endif

    /* Consumes very little CPU, but need highest priority to handle mesh network ACKs */
    //scheduler_add_task(new wirelessTask(PRIORITY_CRITICAL));

    /* Change "#if 0" to "#if 1" to run period tasks; @see period_callbacks.cpp */
    #if 0
    scheduler_add_task(new periodicSchedulerTask());
    #endif

    /* The task for the IR receiver */
    // scheduler_add_task(new remoteTask  (PRIORITY_LOW));

    /* Your tasks should probably used PRIORITY_MEDIUM or PRIORITY_LOW because you want the terminal
     * task to always be responsive so you can poke around in case something goes wrong.
     */

    /**
     * This is a the board demonstration task that can be used to test the board.
     * This also shows you how to send a wireless packets to other boards.
     */
    #if 0
        scheduler_add_task(new example_io_demo());
    #endif

    /**
     * Change "#if 0" to "#if 1" to enable examples.
     * Try these examples one at a time.
     */
    #if 0
        scheduler_add_task(new example_task());
        scheduler_add_task(new example_alarm());
        scheduler_add_task(new example_logger_qset());
        scheduler_add_task(new example_nv_vars());
    #endif

    /**
	 * Try the rx / tx tasks together to see how they queue data to each other.
	 */
    #if 0
        scheduler_add_task(new queue_tx());
        scheduler_add_task(new queue_rx());
    #endif

    /**
     * Another example of shared handles and producer/consumer using a queue.
     * In this example, producer will produce as fast as the consumer can consume.
     */
    #if 0
        scheduler_add_task(new producer());
        scheduler_add_task(new consumer());
    #endif

    /**
     * If you have RN-XV on your board, you can connect to Wifi using this task.
     * This does two things for us:
     *   1.  The task allows us to perform HTTP web requests (@see wifiTask)
     *   2.  Terminal task can accept commands from TCP/IP through Wifly module.
     *
     * To add terminal command channel, add this at terminal.cpp :: taskEntry() function:
     * @code
     *     // Assuming Wifly is on Uart3
     *     addCommandChannel(Uart3::getInstance(), false);
     * @endcode
     */
    #if 0
        Uart3 &u3 = Uart3::getInstance();
        u3.init(WIFI_BAUD_RATE, WIFI_RXQ_SIZE, WIFI_TXQ_SIZE);
        scheduler_add_task(new wifiTask(Uart3::getInstance(), PRIORITY_LOW));
    #endif

    scheduler_start(); ///< This shouldn't return
    return -1;
}
